/* -*- mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL TBRICKS
 * AB BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THISS OFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef ELF_HELPERS_H
#define ELF_HELPERS_H

#include <assert.h>
#include <libelf.h>
#include <stdbool.h>
#include <string.h>

#include "shared/log.h"

bool elf_get_shdr(Elf *elf, Elf_Scn *scn, Elf64_Shdr *shdr)
{
    assert(elf);
    assert(scn);
    assert(shdr);

    size_t ident_size;
    char *elf_ident = elf_getident(elf, &ident_size);
    if (elf_ident == NULL || ident_size <= EI_CLASS) {
        _E("Cannot read ELF ident");
        return false;
    }

    if (elf_ident[EI_CLASS] == ELFCLASS32) {
        Elf32_Shdr *shdr32;
        shdr32 = elf32_getshdr(scn);
        if (shdr32 == NULL) {
            _E("elf32_getshdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }

        shdr->sh_name      = shdr32->sh_name;
        shdr->sh_type      = shdr32->sh_type;
        shdr->sh_flags     = (Elf64_Xword) shdr32->sh_flags;
        shdr->sh_addr      = (Elf64_Addr) shdr32->sh_addr;
        shdr->sh_offset    = (Elf64_Off) shdr32->sh_offset;
        shdr->sh_size      = (Elf64_Xword) shdr32->sh_size;
        shdr->sh_link      = shdr32->sh_link;
        shdr->sh_info      = shdr32->sh_info;
        shdr->sh_addralign = (Elf64_Xword) shdr32->sh_addralign;
        shdr->sh_entsize   = (Elf64_Xword) shdr32->sh_entsize;
    } else {
        Elf64_Shdr *shdr64 = elf64_getshdr(scn);
        if (shdr64 == NULL) {
            _E("elf64_getshdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }
        *shdr = *shdr64;
    }
    return true;
}

bool elf_get_ehdr(Elf *elf, Elf64_Ehdr *ehdr)
{
    assert(elf);
    assert(ehdr);

    size_t ident_size;
    char *elf_ident = elf_getident(elf, &ident_size);
    if (elf_ident == NULL || ident_size <= EI_CLASS) {
        _E("Cannot read ELF ident");
        return false;
    }

    if (elf_ident[EI_CLASS] == ELFCLASS32) {
        Elf32_Ehdr *ehdr32;
        ehdr32 = elf32_getehdr(elf);
        if (ehdr32 == NULL) {
            _E("elf32_getehdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }

        (void) memcpy(ehdr->e_ident,
                      ehdr32->e_ident,
                      sizeof(ehdr32->e_ident));

        ehdr->e_type      = ehdr32->e_type;
        ehdr->e_machine   = ehdr32->e_machine;
        ehdr->e_version   = ehdr32->e_version;
        ehdr->e_entry     = ehdr32->e_entry;
        ehdr->e_phoff     = ehdr32->e_phoff;
        ehdr->e_shoff     = ehdr32->e_shoff;
        ehdr->e_flags     = ehdr32->e_flags;
        ehdr->e_ehsize    = ehdr32->e_ehsize;
        ehdr->e_phentsize = ehdr32->e_phentsize;
        ehdr->e_phnum     = ehdr32->e_phnum;
        ehdr->e_shentsize = ehdr32->e_shentsize;
        ehdr->e_shnum     = ehdr32->e_shnum;
        ehdr->e_shstrndx  = ehdr32->e_shstrndx;
    } else {
        Elf64_Ehdr *ehdr64 = elf64_getehdr(elf);
        if (ehdr64 == NULL) {
            _E("elf64_getehdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }
        *ehdr = *ehdr64;
    }
    return true;
}

bool elf_get_phdr(Elf *elf, int index, Elf64_Phdr *phdr)
{
    assert(elf);
    assert(index >= 0);
    assert(phdr);

    size_t ident_size;
    char *elf_ident = elf_getident(elf, &ident_size);
    if (elf_ident == NULL || ident_size <= EI_CLASS) {
        _E("Cannot read ELF ident");
        return false;
    }

    size_t phdr_num;
    if (elf_getphdrnum(elf, &phdr_num) == -1) {
        _E("Cannot get program headers count");
        return false;
    }

    if ((size_t)index >= phdr_num) {
        _E("Index is larger than the number of program headers");
        return false;
    }

    if (elf_ident[EI_CLASS] == ELFCLASS32) {
        Elf32_Phdr *phdr32;
        phdr32 = elf32_getphdr(elf);
        if (phdr32 == NULL) {
            _E("elf32_getphdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }

        phdr32 += index;

        phdr->p_type   = phdr32->p_type;
        phdr->p_offset = phdr32->p_offset;
        phdr->p_vaddr  = (Elf64_Addr) phdr32->p_vaddr;
        phdr->p_paddr  = (Elf64_Addr) phdr32->p_paddr;
        phdr->p_filesz = (Elf64_Xword) phdr32->p_filesz;
        phdr->p_memsz  = (Elf64_Xword) phdr32->p_memsz;
        phdr->p_flags  = phdr32->p_flags;
        phdr->p_align  = (Elf64_Xword) phdr32->p_align;
    } else {
        Elf64_Phdr *phdr64 = elf64_getphdr(elf);
        if (phdr64 == NULL) {
            _E("elf64_getphdr error: %s", elf_errmsg(elf_errno()));
            return false;
        }
        phdr64 += index;
        *phdr = *phdr64;
    }
    return true;
}

bool elf_get_sym(Elf *elf, Elf_Data *data, int index, Elf64_Sym *sym)
{
    assert(elf);
    assert(index >= 0);
    assert(sym);

    size_t ident_size;
    char *elf_ident = elf_getident(elf, &ident_size);
    if (elf_ident == NULL || ident_size <= EI_CLASS) {
        _E("Cannot read ELF ident");
        return false;
    }

    if (elf_ident[EI_CLASS] == ELFCLASS32) {
        Elf32_Sym *sym32 = &((Elf32_Sym *)data->d_buf)[index];

        sym->st_name  = sym32->st_name;
        sym->st_info  = sym32->st_info;
        sym->st_other = sym32->st_other;
        sym->st_shndx = sym32->st_shndx;
        sym->st_value = sym32->st_value;
        sym->st_size  = sym32->st_size;
    } else {
        *sym = ((Elf64_Sym *)data->d_buf)[index];
    }
    return true;
}
#endif  // ELF_HELPERS_H
